#! /bin/bash # pour emacs

REMOTE_LIP_BACKUP="$REMOTE_LIP_BACKUP --server"

RSYNC_OPTIONS="--archive --cvs-exclude --compress --delete --one-file-system"
RSYNC_OPTIONS="$RSYNC_OPTIONS --delete-excluded"
RSYNC_OPTIONS="$RSYNC_OPTIONS --rsync-path=\"$REMOTE_LIP_BACKUP rsync\""
RSYNC_OPTIONS="$RSYNC_OPTIONS --rsh=\"$SSH\""
#RSYNC_OPTIONS="$RSYNC_OPTIONS --dry-run"

SSH_AGENT_PID=""
SSH_AUTH_SOCK=""

#RSYNC_RSH="$SSH"
#export RSYNC_RSH

}

# script

VERSION=1.0

#if [ -x /soft/eleves/`uname -p`/bin/echo ]; then
#    ECHO=/soft/eleves/`uname -p`/bin/echo
#else
    ECHO=echo 1>&2
#fi

log() { #niveau type message
    (
	if [ "$1" -le $VERBOSE_LEVEL ]; then
	    type="$2"
	    fonction="$3"
	    shift
	    shift
	    shift
	    [ $SERVER_MODE = 1 ] && $ECHO -n "S: "
	    [ $SERVER_MODE != 1 ] && $ECHO -n "C: "
	    case $type in
		0) $ECHO -n "--> ";;
		1) $ECHO -n "<-- ";;
		2) $ECHO -n "    ";;
	    esac
	    $ECHO -n $fonction
	    if [ ! -z "$*" ]; then
		$ECHO ": $*"
	    else
		$ECHO
	    fi
	fi
    ) 1>&2
}

usage() {
    log 2 0 "usage"
    echo "Usage: `basename $0` [option...] action [actions...]"
    echo "  Using configuration file ${CONFFILE}"
    echo "  Options:"
    echo "   --comp type : compression to use (gzip, bzip2 or none)"
    echo "   -v|--verbose : incremente verbosity"
    echo "   -q|--quiet : decrement verbosity"
    echo "   -h|--help : this help"
    echo "   --force : do not check permissions (for experts only)"
    echo "  Actions:"
    echo "   --system : save the system state"
    echo "   --dir directory : save /directory"
    echo "   --user [name] : save $BASEHOME/name "'(`logname` if name not specified)'
    log 2 1 "usage"
}

set_compression() {
    log 2 0 "set_compression" "$COMPRESSION"
    case "$COMPRESSION" in
	gzip)
	    COMP="--gzip"
	    COMP_EXT=".gz"
	    COMP_PROG=gzip
	    COMP_LEVEL=-9
	    ;;
	bzip2)
	    COMP="--bzip2"
	    COMP_EXT=".bz2"
	    COMP_PROG=bzip2
	    COMP_LEVEL=-9
	    ;;
	none)
	    COMP=""
	    COMP_EXT=""
	    COMP_PROG=cat
	    COMP_LEVEL=
	    ;;		    
	*)
	    log 0 2 set_compression "unknown compression type $COMPRESSION. Using gzip"
	    COMPRESSION=gzip
	    set_compression
	    ;;
    esac
    log 2 1 "set_compression"
}

#RSYNC="rsync --rsh=/home/vdanjean/bin/ssh-rsync"

# -aCz --delete
#      /home/mquinson info:/home/mquinson/save/`hostname`

identity() { #commande
    log 2 0 "identity" "$@"
    if [ `id -u` = 0 -a ! -z "$SSH_SENDER_INSTEAD_OF_ROOT" ]; then
	log 2 2 "identity" \
	    "executing as $SSH_SENDER_INSTEAD_OF_ROOT instead of root"
	su $SSH_SENDER_INSTEAD_OF_ROOT -c "$*"
    else
	"$@"
    fi
    log 2 1 "identity"
}

send(){ # file ext
    log 2 0 "send" "(file: $1 ext: $2)"
    identity $SSH $ARCHIVE_HOSTNAME \
	$REMOTE_LIP_BACKUP send $ARCHIVE_PATH $1 $2 $MAX_ROTATE
    log 2 1 "send"
}

send_dpkg() {
    log 1 0 "send_dpkg"
    dpkg --get-selections \* | ${COMP_PROG} ${COMP_LEVEL} | \
	send dpkg.getselection "${COMP_EXT}"
    log 1 1 "send_dpkg"
}

send_dir() { # basedir dir [--rsync] [--exclude file1 file2 ...]
    log 1 0 "send_dir" "$*"
    basedir="$1"
    shift
    dir=`echo "$1" | sed -e 's|^/\(.*\)$|\1|' -e 's|^\(.*\)/$|\1|'`
    shift
    if [ "$1" = --rsync ]; then
	methode=rsync
    else
	methode=tar
    fi
    if [ ! \( -f "$basedir/$dir" -o -d "$basedir/$dir" \) ]; then
	echo "$basedir/$dir" do not exists. Not saving this one...
	return 1
    fi
    filename=`echo $dir | sed -e 's/\//_/g'`
    log 1 2 "send_dir" "finding directories to exclude"
    EXCL=""
    if [ -f "$basedir/$dir/.nosave-list" ]; then
	log 2 2 "send_dir" "reading $basedir/$dir/.nosave-list"
	EXCL=`cat "$basedir/$dir/.nosave-list" | \
	    sed -e 's|^-*/*\([^/# +].*\)/* *$|--exclude '"$dir/"'\1|p;d'`
    fi
    EXCL="$EXCL "`(cd "$basedir" && find $dir -xdev -name ".nosave" -type f -print -a -prune | \
		sed -e 's/^\(.*\)\/\.nosave$/--exclude \1/' )`
    NOSAVE=`echo $EXCL | sed -e 's/--exclude //g'`
    NOLIST=""
    if [ "$1" = "--exclude" ]; then
	shift
	while [ $# != 0 ]; do
	    if [ -e "$basedir/$dir/$1" ]; then
		EXCL="$EXCL --exclude $dir/$1"
		NOSAVE="$NOSAVE $dir/$1"
		NOLIST="$NOLIST -path $dir/$1 -prune -o"
	    fi
	    shift
	done
    fi
    log 3 2 "send_dir" "excluding $EXCL"
    log 1 2 "send_dir" "sending"
    case $methode in
	tar)
	    (cd "$basedir" && \
		tar cf - ${COMP} --one-file-system $EXCL $dir) \
		| send $filename .tar${COMP_EXT}
	    ;;
	rsync)
	    EXCL_RSYNC=`echo " $EXCL " | \
			sed -e "s|--exclude $dir|--exclude |g"`
	    identity $SSH "$ARCHIVE_HOSTNAME" \
		$REMOTE_LIP_BACKUP mkdir "$ARCHIVE_PATH"/home/$dir
	    eval identity `echo "$RSYNC"` $RSYNC_VERB $RSYNC_OPTIONS $EXCL_RSYNC \
		"$basedir/$dir/" \
		"$ARCHIVE_HOSTNAME":"$ARCHIVE_PATH"/home/$dir
	    ;;
    esac
    if [ "$EXCL" != " " ]; then
	log 1 2 "send_dir" "list file"
	EXCL=""
	if [ -f "$basedir/$dir/.nolist-list" ]; then
	    log 2 2 "send_dir" "reading $basedir/$dir/.nolist-list"
	    EXCL=`cat "$basedir/$dir/.nolist-list" | \
	        sed -e 's|^-*/*\([^/# +].*\)/* *$| -path '"$dir/"'\1 -prune -o |p;d'`
	fi
	EXCL="$EXCL "`(cd "$basedir" && find $dir -xdev -name ".nolist" -type f | \
			sed -e 's/^\(.*\)\/\.nolist$/ -path \1 -prune -o/') | \
			sort`
	if [ ! -z "$EXCL" ]; then
	    log 3 2 "send_dir" "Not Printing :"
	    log 3 2 "send_dir" "$EXCL"
	fi
	cd $basedir && \
	    find $NOSAVE $NOLIST $EXCL -xdev -exec ls -ld {} \; | \
	    ${COMP_PROG} ${COMP_LEVEL} | send $filename .lst${COMP_EXT}
    fi
    log 1 1 "send_dir"
}

send_diversion() {
    log 1 0 "send_diversion"
    args=`LC_ALL=C /usr/sbin/dpkg-divert --local --list|
	    grep "local diversion" |
	    sed 's/local diversion of //'|
	    sed 's/ to .*$//'`
    if [ -z "$args" ]; then
	log 1 2 "send_diversion" "nothing to send"
    else
	tar cf - ${COMP} $args | send dpkg-divert .tar${COMP_EXT}
    fi
    log 1 1 "send_diversion"
}    

send_system() {
    log 1 0 "send_system"
    if [ `id -u` != 0 ]; then
	if [ "$FORCE" != yes ];then
	    log 0 2 "send_system" "Only root can do that" \
		"(or use --force if you know what you do)"
	    log 1 1 "send_system" 
	    return 1
	else
	    log 0 2 "send_system" "Warning: you are not root"
	fi
    fi
    [ "$SEND_DPKG" = 1 ] && send_dpkg
    [ "$SEND_DIVERTED" = 1 ] && send_diversion
    
    for dir in $SEND_DIRS ""; do
	[ -z "$dir" ] && continue
	base=`eval echo '$SEND_'$dir'_DIR'`
	base=`echo "$base" | sed -e 's|^/*\([^/].*[^/]\)/*$|\1|'`
	excl=`eval echo '$SEND_'$dir'_EXCLUDE'`
	send_dir / "$base" ${excl:+--exclude $excl}
    done
    log 1 1 "send_system" 
}

send_user() { # login
    log 1 0 "send_user" "$1"
    if [ ! -d $BASEHOME/$1 ]; then
	log 0 2 "send_user" "$1 has no home directory $BASEHOME/$1"
	log 1 1 "send_user"
	return 1
    fi
    if [ `logname` != "$1" ]; then
	if [ "$FORCE" != yes ]; then
	    log 0 2 "send_user" "Only the user $1 can save its HOME" \
		"(or use --force if you know what you do)"
	    log 1 1 "send_system" 
	    return 1
	else
	    log 0 2 "send_user" "Warning: `logname` saving $1's HOME"
	fi
    fi
    send_dir $BASEHOME $1 --rsync
	#send_dir /home $1
    log 1 1 "send_user"
}

main() {
    log 2 0 "main" "$@"
    LIST_USER=""
    LIST_DIR=""
    SYSTEM=""
    RSYNC_VERB=""
    if [ -z "$ARCHIVE_HOSTNAME" ]; then
	log 0 2 "error" "ARCHIVE_HOSTNAME not set"
	usage
	exit 1
    fi
    if [ -z "$ARCHIVE_PATH" ]; then
	log 0 2 "error" "ARCHIVE_PATH not set"
	usage
	exit 1
    fi
    if [ -z "$ARCHIVE_USER" ]; then
	log 0 2 "error" "ARCHIVE_USER not set"
	usage
	exit 1
    fi
    if [ "$SSH_KEY" -a ! -f "$SSH_KEY" ]; then
	log 0 2 "error" "ssh key $SSH_KEY does not exist"
	exit 1
    fi
    if [ "$TEST_REACHING" ]; then
	if $TEST_REACHING >/dev/null 2>&1; then
	    log 2 2 "main" "Reaching test succeed"
	else
	    log 0 2 "error" "Reaching test failed"
	    log 0 2 "error" "Test used: $TEST_REACHING"
	    exit 1
	fi
    fi
    while [ $# != 0 ]; do
	case "$1" in
	    --comp) COMPRESSION="$2" ; shift ;;
	    --dir) LIST_DIR="$LIST_DIR${2:+ $2}" ; shift ;;
	    --system) SYSTEM=1 ;;
	    --force) FORCE=yes ;;
	    --user) LIST_USER="$LIST_USER ${2:-`logname`}" ; shift ;;
	    -v|--verbose) VERBOSE_LEVEL=`expr $VERBOSE_LEVEL + 1`
		RSYNC_VERB="$RSYNC_VERB -v"
		REMOTE_LIP_BACKUP="$REMOTE_LIP_BACKUP -v" ;;
	    -q|--quiet) VERBOSE_LEVEL=`expr $VERBOSE_LEVEL - 1`
		RSYNC_VERB="$RSYNC_VERB -q" 
		REMOTE_LIP_BACKUP="$REMOTE_LIP_BACKUP -q" ;;
	    -h|--help) usage ; exit 0;;
	    -*) echo "Unknown option $1"
		usage ; exit 1;;
	    *)  echo "Unknown argument $1"
		usage ; exit 1;;
	esac
	shift
    done
    
    if [ -z "${SYSTEM}${LIST_DIR}${LIST_USER}" ]; then
	usage ; exit 1
    fi
    
    set_compression
    
    VER=`identity $SSH $ARCHIVE_HOSTNAME \
	$REMOTE_LIP_BACKUP version`
    if [ "$VER" != "$VERSION" ]; then
	log 0 2 "main" "client/server versions mismatch : $VERSION/$VER"
	exit 1
    fi

    if [ "$SYSTEM" = 1 ]; then
	send_system
    fi
    
    if [ ! -z "$LIST_DIR" ]; then
	for dir in $LIST_DIR; do
		send_dir / "$dir"
	done
    fi

    if [ ! -z "$LIST_USER" ]; then
	for user in $LIST_USER; do
		send_user "$user"
	done
    fi
    log 2 1 "main" "$@"
}

server_rotate() { # file rot_max
    log 2 0 "server_rotate" "$@"
    file="$1"
    if [ -z "$2" ]; then
	rot_max="$ROTATE_MAX"
    else
	rot_max="$2"
    fi
    ext="0"
    while [ -f "$1".$ext -a $ext -lt $rot_max ]; do
	ext=`expr $ext + 1`
    done
    liaison="."
    while [ "$ext" != "" ]; do
	prev=`expr $ext - 1`
	if [ $prev = -1 ]; then
	    prev=""
	    liaison=""
	fi
	if [ -f $file$liaison$prev ]; then
	    mv $file$liaison$prev $file.$ext
	fi
	ext=$prev
    done
    log 2 1 "server_rotate" "$@"
}

server_read_file() { # savedir file ext nb_rot
    log 2 0 "server_read_file" "$@"
    mkdir -p $1
    FILE="$1/$2$3"

    cat > "$FILE".tmp
    if [ ! -f "$FILE" ]; then
        mv "$FILE".tmp "$FILE"
    else
        if zcmp -s "$FILE" "$FILE".tmp ; then
            touch "$FILE"
            rm -f "$FILE".tmp
        else
            server_rotate "$FILE" "$4"
            mv "$FILE".tmp "$FILE"
	fi
    fi
    log 2 1 "server_read_file" "$@"
}

do_server_exec() {
    log 2 0 "do_serveur_exec" "$@"

    while [ "$1" = "-v" -o "$1" = "-q" ]; do
	if [ "$1" = "-v" ]; then
	    VERBOSE_LEVEL=`expr $VERBOSE_LEVEL + 1`
	else
	    VERBOSE_LEVEL=`expr $VERBOSE_LEVEL - 1`
	fi
	shift
    done

    case "$1" in
	"send"*)
	    shift;
	    server_read_file "$@";;
	"rsync"*)
	    shift;
	    rsync "$@";;
	"mkdir"*)
	    shift;
	    mkdir -p "$1";;
	"version")
	    echo "$VERSION";;
	"put_ssh_key")
	    ;;
	*)
	    log 0 2 "do_serveur_exec" \
		"Error : invalid command $SSH_ORIGINAL_COMMAND"
	;;
    esac
    log 2 1 "do_serveur_exec" "$@"
}

do_serveur() {
    log 2 0 "do_serveur" "$@"
    if [ -z "$SSH_ORIGINAL_COMMAND" ]; then
	do_server_exec "$@"
	log 2 1 "do_serveur" "$@"
	exit 0
    fi
    log 2 0 "do_serveur" "using $SSH_ORIGINAL_COMMAND"
    case "$SSH_ORIGINAL_COMMAND" in
	rsync*)
	    log 0 2 "do_serveur" "Error : perms rsync"
	    $SSH_ORIGINAL_COMMAND
	    ;;
	"$REMOTE_LIP_BACKUP "*)
	    set -- $SSH_ORIGINAL_COMMAND
	    shift ; shift # lip-backup --server
	    do_server_exec "$@"
	    ;;
	"mkdir -p"*)
	    log 0 2 "do_serveur" "Error : perms rsync"
	    $SSH_ORIGINAL_COMMAND
	    ;;
	*)
	    log 0 2 "do_serveur" "Error : invalid command $SSH_ORIGINAL_COMMAND"
	    ;;
    esac
    log 2 1 "do_serveur" "$@"
}

case "$1" in
    --server)
	SERVER_MODE=1
	log 2 2 "start" "Server mode on `uname -n`"
	shift
	do_serveur "$@" ;;
    *)
	SERVER_MODE=0
	set_config
	log 2 2 "start" "Client mode on `uname -n`"
	log 2 2 "start" "args: $@"
	main "$@" ;;
esac
